import threading
import django
from django.apps import AppConfig, apps
from django.utils import six

from . import settings


class TenantAwareThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        from django.db import connection
        self.tenant = connection.get_tenant()
        super(TenantAwareThread, self).__init__(*args, **kwargs)

    def run(self):
        from django.db import connection
        old_tenant = connection.get_tenant()
        connection.set_tenant(self.tenant)
        try:
            super(TenantAwareThread, self).run()
        finally:
            connection.set_tenant(old_tenant)
            connection.close()


class _Timer(TenantAwareThread):
    """Call a function after a specified number of seconds:

            t = Timer(30.0, f, args=[], kwargs={})
            t.start()
            t.cancel()     # stop the timer's action if it's still waiting

    """

    def __init__(self, interval, function, args=[], kwargs={}):
        super(_Timer, self).__init__()
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = threading.Event()

    def cancel(self):
        """Stop the timer if it hasn't finished yet"""
        self.finished.set()

    def run(self):
        self.finished.wait(self.interval)
        if not self.finished.is_set():
            self.function(*self.args, **self.kwargs)
        self.finished.set()


class _MainThread(TenantAwareThread):

    def __init__(self):
        if six.PY3:
            super(_MainThread, self).__init__(name="MainThread", daemon=False)
            self._set_tstate_lock()
            self._started.set()
            self._set_ident()
            with threading._active_limbo_lock:
                _active[self._ident] = self
        else:
            super(_MainThread, self).__init__(name="MainThread")
            self._Thread__started.set()
            self._set_ident()
            with threading._active_limbo_lock:
                threading._active[threading._get_ident()] = self

    def _set_daemon(self):
        return False

    def _exitfunc(self):
        self._Thread__stop()
        t = threading._pickSomeNonDaemonThread()
        if t:
            if __debug__:
                self._note("%s: waiting for other threads", self)
        while t:
            t.join()
            t = threading._pickSomeNonDaemonThread()
        if __debug__:
            self._note("%s: exiting", self)
        self._Thread__delete()


class _DummyThread(TenantAwareThread):

    def __init__(self):
        if six.PY3:
            super(_DummyThread, self).__init__(name=threading._newname("Dummy-%d"), daemon=True)
            self._started.set()
            self._set_ident()
            with threading._active_limbo_lock:
                threading._active[self._ident] = self
            return

        super(_DummyThread, self).__init__(name=threading._newname("Dummy-%d"))

        # Thread.__block consumes an OS-level locking primitive, which
        # can never be used by a _DummyThread.  Since a _DummyThread
        # instance is immortal, that's bad, so release this resource.
        del self._Thread__block

        self._Thread__started.set()
        self._set_ident()
        with threading._active_limbo_lock:
            threading._active[threading._get_ident()] = self

    def _set_daemon(self):
        return True

    def join(self, timeout=None):
        assert False, "cannot join a dummy thread"


class MultitenantAppConfig(AppConfig):
    name = 'hobo.multitenant'
    verbose_name = 'Multitenant'

    def ready(self):
        from django.db.migrations import operations
        from django.db import migrations
        from django import conf

        # Install tenant aware settings
        if not isinstance(conf.settings._wrapped,
                          settings.TenantSettingsWrapper):
            conf.settings._wrapped = settings.TenantSettingsWrapper(
                conf.settings._wrapped)
            # reset settings getattr method to a cache-less version, to cancel
            # https://code.djangoproject.com/ticket/27625.
            conf.LazySettings.__getattr__ = lambda self, name: getattr(self._wrapped, name)

        # Install tenant aware Thread class
        threading.Thread = TenantAwareThread
        threading._DummyThread = _DummyThread
        threading._MainThread = _MainThread
        threading._Timer = _Timer

        if (1, 7) <= django.VERSION < (1, 8):
            class RunPythonOverride(operations.RunPython):
                def can_migrate(self, app_label, schema_editor):
                    app = apps.get_app_config(app_label)
                    for model in app.get_models():
                        if self.allowed_to_migrate(schema_editor.connection.alias, model):
                            return True
                    return False

                def database_forwards(self, app_label, schema_editor, *args, **kwargs):
                    if not self.can_migrate(app_label, schema_editor):
                        return
                    super(RunPythonOverride, self).database_forwards(app_label, schema_editor, *args, **kwargs)

                def database_backwards(self, app_label, schema_editor, *args, **kwargs):
                    if not self.can_migrate(app_label, schema_editor):
                        return
                    super(RunPythonOverride, self).database_backwards(app_label, schema_editor, *args, **kwargs)

            operations.RunPython = RunPythonOverride
            migrations.RunPython = RunPythonOverride
