import hashlib
from importlib import import_module
import threading
import time

from django.db import connection
from django.conf import UserSettingsHolder


def import_class(klass):
    module, attr = klass.rsplit('.', 1)
    return getattr(import_module(module), attr)


class TenantSettingsWrapper(object):
    local = threading.local()

    def __init__(self, default_settings):
        self.__dict__['tenants_settings'] = {}
        self.__dict__['default_settings'] = default_settings

    def clear_tenants_settings(self):
        self.__dict__['tenants_settings'] = {}

    def load_tenant_settings(self, wrapped, tenant, tenant_settings,
                             last_time):
        '''Load tenant settings from loaders into tenant_settings object, only
           if any of the loaders say it is more recent than last update time'''
        settings = self.default_settings
        update_time = time.time()
        if last_time and update_time - last_time < 3:
            return tenant_settings, last_time
        for loader in getattr(settings, 'TENANT_SETTINGS_LOADERS', []):
            loader_class = import_class(loader)
            loader_instance = loader_class()
            new_time = loader_instance.get_new_time(tenant)
            if (not new_time and last_time) \
                    or (new_time and not last_time) \
                    or (new_time and new_time > last_time):
                # something is new, call all loaders
                for loader in settings.TENANT_SETTINGS_LOADERS:
                    loader_class = import_class(loader)
                    loader_instance = loader_class()
                    loader_instance.update_settings(tenant_settings, tenant)
                return tenant_settings, update_time
        return tenant_settings, last_time

    def get_tenant_settings(self, wrapped, tenant):
        '''Get last loaded settings for tenant, try to update it by loading
           settings again is last loading time is less recent thant settings
           data store. Compare with last modification time is done in the
           load_tenant_settings() method.
        '''
        tenant_settings, last_time = self.tenants_settings.get(
            tenant.domain_url, (None, None))
        if tenant_settings is None:
            tenant_settings = UserSettingsHolder(wrapped)
        tenant_settings, last_time = self.load_tenant_settings(
            wrapped, tenant, tenant_settings, last_time)
        self.tenants_settings[tenant.domain_url] = tenant_settings, last_time
        return tenant_settings

    def get_wrapped(self):
        if getattr(self.local, 'in_get_wrapped', False):
            return self.default_settings
        try:
            self.local.in_get_wrapped = True
            tenant = connection.get_tenant()
            if not hasattr(tenant, 'domain_url'):
                return self.default_settings
            return self.get_tenant_settings(self.default_settings, tenant)
        finally:
            self.local.in_get_wrapped = False

    def __getattr__(self, name):
        return getattr(self.get_wrapped(), name)

    def __setattr__(self, name, value):
        setattr(self.get_wrapped(), name, value)

    def __delattr__(self, name):
        delattr(self.get_wrapped(), name)
